
#import "GBDatabaseFileDownloadController.h"
#import "GBSourceListDataSource.h"
#import "GBSourceListItem.h"
#import "GameBaseAppDelegate.h"
#import "NSString+PathUtilities.h"
#import "zlib.h"


static NSString* GBDatabasesURLString = @"http://www.sidmusic.org/gamebase64browser/gb64_databases.txt";

static NSString* GBURLRequestUserAgentString = nil;

@implementation GBDatabaseFileDownloadController


@synthesize databaseFolderName = _databaseFolderName;


// ----------------------------------------------------------------------------
- (void) awakeFromNib
// ----------------------------------------------------------------------------
{
	NSDictionary* infoDictionary = [[NSBundle mainBundle] infoDictionary];
	NSString* appNameString = [infoDictionary objectForKey:@"CFBundleName"];
	NSString* appVersionString = [infoDictionary objectForKey:@"CFBundleVersion"];
	NSString* osVersionString = [[NSProcessInfo processInfo] operatingSystemVersionString];
	
	GBURLRequestUserAgentString = [NSString stringWithFormat:@"%@/%@ (Mac OS X, %@)", appNameString, appVersionString, osVersionString];
	
	databaseFileURLs = [[NSMutableArray alloc] init];
	downloadDatabaseItem = nil;
	_databaseFolderName = nil;
	
	[downloadButton setEnabled:NO];
	[downloadVersionPopUpButton setEnabled:NO];
	
	NSURL* url = [NSURL URLWithString:GBDatabasesURLString];
	NSMutableURLRequest* request = [NSMutableURLRequest requestWithURL:url cachePolicy:NSURLRequestReloadIgnoringLocalCacheData timeoutInterval:60.0];
	[request setValue:GBURLRequestUserAgentString forHTTPHeaderField:@"User-Agent"];
	
	databaseListURLConnection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
	if (databaseListURLConnection != nil)
		databaseListData = [NSMutableData data];
}


// ----------------------------------------------------------------------------
- (NSString*) detectDatabaseVersionOfFolder:(NSString*)databaseStoragePath
// ----------------------------------------------------------------------------
{
	// Check for a GBC_v0x.mdb file to determine the version of Gamebase64
	NSString* foundVersionString = nil;
	NSArray* databaseFiles = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:databaseStoragePath error:NULL];
	for (__strong NSString* databaseFile in databaseFiles)
	{
		databaseFile = [databaseFile lowercaseString];
		if ([[databaseFile pathExtension] isEqualToString:@"mdb"])
		{
			foundVersionString = [[databaseFile stringByDeletingPathExtension] stringByRemovingPrefix:@"gbc_"];
			break;
		}
		else if ([[databaseFile pathExtension] isEqualToString:@"exe"])
		{
			foundVersionString = [[databaseFile stringByDeletingPathExtension] stringByRemovingPrefix:@"gb64"];
			break;
		}
	}
	
	return foundVersionString;
}


// ----------------------------------------------------------------------------
- (void) prepareDownloadForDatabase:(GBSourceListItem*)databaseItem
// ----------------------------------------------------------------------------
{
	if (databaseFileURLs == nil)
		return;
	
	downloadDatabaseItem = databaseItem;
	self.databaseFolderName = [[databaseItem path] lastPathComponent];
	
	NSString* databaseStoragePath = [databaseItem path];
	NSString* foundVersionString = [self detectDatabaseVersionOfFolder:databaseStoragePath];
	
	NSMenu* menu = [[NSMenu alloc] initWithTitle:@""];
	BOOL detectedVersion = NO;
	
	NSInteger index = 0;
	for (NSMutableDictionary* databaseFileURL in databaseFileURLs)
	{
		NSString* versionString = [databaseFileURL objectForKey:@"Version"];
		NSString* menuItemString = [NSString stringWithFormat:@"GameBase64 %@", versionString];
		NSMenuItem* menuItem = [[NSMenuItem alloc] initWithTitle:menuItemString action:nil keyEquivalent:@""];
		[menuItem setTag:index];
		[menuItem setTarget:self];
		[menu addItem:menuItem];
		
		// If there was a .mdb file already in the folder, select that version as a default
		if (foundVersionString != nil && [versionString caseInsensitiveCompare:foundVersionString] == NSOrderedSame)
		{
			[menuItem setState:NSOnState];
			detectedVersion = YES;
		}
		
		index++;
	}

	if (detectedVersion)
		[downloadVersionDetectedLabel setStringValue:[NSString stringWithFormat:@"This folder seems to contain GameBase64 %@", foundVersionString]];
	
	[downloadVersionSelectLabel setHidden:detectedVersion];
	[downloadVersionPopUpButton setHidden:detectedVersion];
	[downloadVersionDetectedLabel setHidden:!detectedVersion];
	
	[downloadVersionPopUpButton setMenu:menu];
	[downloadVersionPopUpButton setEnabled:YES];
}


// ----------------------------------------------------------------------------
- (IBAction) downloadButtonClicked:(id) sender
// ----------------------------------------------------------------------------
{
	[downloadVersionPopUpButton setEnabled:NO];
    [downloadButton setEnabled:NO];
    
	NSInteger databaseIndex = [downloadVersionPopUpButton selectedTag];
	
	NSMutableDictionary* databaseFileURL = [databaseFileURLs objectAtIndex:databaseIndex];
	
	NSString* urlString = [databaseFileURL objectForKey:@"DatabaseURL"];
	databaseFileName = [[urlString lastPathComponent] stringByDeletingPathExtension];
	
	NSURL* url = [NSURL URLWithString:urlString];
	NSMutableURLRequest* request = [NSMutableURLRequest requestWithURL:url cachePolicy:NSURLRequestReloadIgnoringLocalCacheData timeoutInterval:60.0];
	[request setValue:GBURLRequestUserAgentString forHTTPHeaderField:@"User-Agent"];
	
	databaseFileURLConnection = [[NSURLConnection alloc] initWithRequest:request delegate:self];
	if (databaseFileURLConnection != nil)
		databaseFileData = [NSMutableData data];
}


#pragma mark -
#pragma mark NSURLConnection delegate methods


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
// ----------------------------------------------------------------------------
{
	if (connection == databaseFileURLConnection)
	{
		downloadFileSize = [response expectedContentLength];
		[databaseFileData setLength:0];
	}
	else
		[databaseListData setLength:0];
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data
// ----------------------------------------------------------------------------
{
	if (connection == databaseFileURLConnection)
	{
		[databaseFileData appendData:data];
		float downloadProgress = 100.0f * [databaseFileData length] / downloadFileSize;
		[downloadProgressIndicator setDoubleValue:downloadProgress];
		float totalSizeinMB = downloadFileSize / 1024.0f / 1024.0f;
		float currentSizeinMB = [databaseFileData length] / 1024.0f / 1024.0f;
		NSString* progressString = [NSString stringWithFormat:@"%.2f MB of %.2f MB", currentSizeinMB, totalSizeinMB];
		[downloadProgressLabel setStringValue:progressString];
	}
	else
		[databaseListData appendData:data];
}


// ----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error
// ----------------------------------------------------------------------------
{
	if (connection == databaseListURLConnection)
	{
		[databaseFileURLs removeAllObjects];
		[downloadButton setEnabled:NO];
		[downloadVersionPopUpButton setEnabled:NO];
	}
}


// ----------------------------------------------------------------------------
- (void) connectionDidFinishLoading:(NSURLConnection*)connection
// ----------------------------------------------------------------------------
{
	if (connection == databaseFileURLConnection)
	{
		BOOL error = NO;
		
		if ([databaseFileData length] > 0)
		{
			NSData* uncompressedData = [databaseFileData gzipInflate];
			if (uncompressedData != nil)
			{
				NSString* databaseFileDestinationPath = [[downloadDatabaseItem path] stringByAppendingPathComponent:databaseFileName];
				[uncompressedData writeToFile:databaseFileDestinationPath atomically:YES];
				
				GameBaseAppDelegate* gamebaseAppDelegate = (GameBaseAppDelegate*)[NSApp delegate];
				[[gamebaseAppDelegate sourceListDataSource] updateDatabaseItem:downloadDatabaseItem asExisting:YES andValid:YES];
				[gamebaseAppDelegate loadDatabase:downloadDatabaseItem];
                
                [downloadButton setEnabled:YES];
                [downloadVersionPopUpButton setEnabled:YES];
			}
			else
				error = YES;
		}
		else
			error = YES;

		databaseFileData = nil;
		databaseFileURLConnection = nil;

		if (error)
		{
            NSAlert* alert = [[NSAlert alloc] init];
            [alert addButtonWithTitle:@"OK"];
            [alert setMessageText:@"Download of database file failed!"];
            [alert setInformativeText:@"Please check your internet connection or try again later."];
            [alert setAlertStyle:NSWarningAlertStyle];
            [alert runModal];
		}
	}
	else
	{
        
        if (databaseListData == nil)
			return;
		
		NSString* databaseListString = [[NSString alloc] initWithData:databaseListData encoding:NSASCIIStringEncoding];
		
		if (databaseListString == nil)
			return;
		
		NSArray* databaseEntries = [databaseListString componentsSeparatedByString:@"\n"];
		
		[databaseFileURLs removeAllObjects];
		
		for (NSString* databaseEntry in databaseEntries)
		{
			if ([databaseEntry length] == 0 || [databaseEntry characterAtIndex:0] == '#')
				continue;
			
			NSArray* databaseEntryComponents = [databaseEntry componentsSeparatedByString:@"|"];
			NSString* versionString = [databaseEntryComponents objectAtIndex:0];
			NSString* urlString = [databaseEntryComponents objectAtIndex:1];
			
			NSMutableDictionary* databaseVersionURLPair = [NSMutableDictionary dictionaryWithCapacity:2];
			[databaseVersionURLPair setObject:versionString forKey:@"Version"];
			[databaseVersionURLPair setObject:urlString forKey:@"DatabaseURL"];
			
			[databaseFileURLs addObject:databaseVersionURLPair];
		}
		
		BOOL downloadURLsAvailable = [databaseFileURLs count] > 0;
		[downloadButton setEnabled:downloadURLsAvailable];
		[downloadVersionPopUpButton setEnabled:downloadURLsAvailable];
		
		databaseListData = nil;
		databaseListURLConnection = nil;
	}
}


@end


#pragma mark -
#pragma mark NSData gzip category


@implementation NSData (NSDataExtension)


// ----------------------------------------------------------------------------
- (NSData*) gzipInflate
// ----------------------------------------------------------------------------
{
	if ([self length] == 0) return self;
	
	unsigned full_length = [self length];
	unsigned half_length = [self length] / 2;
	
	NSMutableData* decompressed = [NSMutableData dataWithLength: full_length + half_length];
	BOOL done = NO;
	int status;
	
	z_stream strm;
	strm.next_in = (Bytef *)[self bytes];
	strm.avail_in = [self length];
	strm.total_out = 0;
	strm.zalloc = Z_NULL;
	strm.zfree = Z_NULL;
	
	if (inflateInit2(&strm, (15+32)) != Z_OK) return nil;
	while (!done)
	{
		// Make sure we have enough room and reset the lengths.
		if (strm.total_out >= [decompressed length])
			[decompressed increaseLengthBy: half_length];
		strm.next_out = ((Bytef*)[decompressed mutableBytes]) + strm.total_out;
		strm.avail_out = [decompressed length] - strm.total_out;
		
		// Inflate another chunk.
		status = inflate (&strm, Z_SYNC_FLUSH);
		if (status == Z_STREAM_END) done = YES;
			else if (status != Z_OK) break;
	}
	if (inflateEnd (&strm) != Z_OK) return nil;
	
	// Set real length.
	if (done)
	{
		[decompressed setLength: strm.total_out];
		return [NSData dataWithData: decompressed];
	}
	else return nil;
}

@end
