#import "GameBaseAppDelegate.h"
#import "GBGameTableDataSource.h"
#import "GBPreferencesController.h"
#import "GBSourceListDataSource.h"
#import "GBSourceListItem.h"
#import "GBImageBrowserDataSource.h"
#import "GBGameDetailController.h"
#import "GBPreferencesWindowController.h"
#import "NSString+PathUtilities.h"
#import "GBGameCollection.h"
#import "GBSourceListItem.h"


@implementation GameBaseAppDelegate

@synthesize gameTableLoaded = _gameTableLoaded;
@synthesize databasePath = _databasePath;
@synthesize databaseConnector = _databaseConnector;
@synthesize window = _window;
@synthesize sourceListDataSource = _sourceListDataSource;
@synthesize loadingProgressIndicator = _loadingProgressIndicator;
@synthesize gameDetailController = _gameDetailController;


// ----------------------------------------------------------------------------
- (void) awakeFromNib
// ----------------------------------------------------------------------------
{
	[[GBPreferencesController sharedInstance] load];

	if (prefsWindowController == nil)
		prefsWindowController = [[GBPreferencesWindowController alloc] init];
	
	[_loadingProgressIndicator setHidden:YES];
	[_loadingProgressIndicator setIndeterminate:YES];
	[loadingLabel setHidden:YES];
}


// ----------------------------------------------------------------------------
- (void) applicationWillTerminate:(NSNotification*)aNotification
// ----------------------------------------------------------------------------
{
	[[GBPreferencesController sharedInstance] save];
}


// ----------------------------------------------------------------------------
- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
// ----------------------------------------------------------------------------
{
	[[GBPreferencesController sharedInstance] load];

	if (gPreferences.mImageGridMode)
		[self switchToImageGridMode:self];
	
	_databaseConnector = [[SQLite3Connector alloc] init];  
	_databasePath = nil;
	databaseStoragePath = nil;
	
	_gameTableLoaded = NO;
	
	GBSourceListItem* databaseItem = [_sourceListDataSource currentDatabaseItem];
	if (databaseItem != nil)
		[self loadDatabase:databaseItem];
	else
		[_sourceListDataSource showDatabaseGuide:YES];
}
	

static NSString* sApplicationSupportPath = nil;

// ----------------------------------------------------------------------------
+ (NSString*) applicationSupportPath
// ----------------------------------------------------------------------------
{
	if (sApplicationSupportPath == nil)
	{
        NSArray*paths = NSSearchPathForDirectoriesInDomains(NSApplicationSupportDirectory, NSUserDomainMask, YES);
        sApplicationSupportPath = [paths firstObject];
        sApplicationSupportPath = [sApplicationSupportPath stringByAppendingPathComponent:@"GameBase64 Browser"];

        BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:sApplicationSupportPath];
        if (!exists)
            [[NSFileManager defaultManager] createDirectoryAtPath:sApplicationSupportPath withIntermediateDirectories:YES attributes:nil error:NULL];
	}
	
	return sApplicationSupportPath;
}


// ----------------------------------------------------------------------------
+ (BOOL) containsValidDatabaseAtPath:(NSString*)storagePath
// ----------------------------------------------------------------------------
{
	NSArray* databaseFiles = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:storagePath error:NULL];
	for (NSString* databaseFile in databaseFiles)
	{
		if ([[databaseFile pathExtension] caseInsensitiveCompare:@"sqlitedb"] == NSOrderedSame)
			return YES;
	}

	return NO;
}


// ----------------------------------------------------------------------------
+ (NSString*) databasePathFromDatabaseStoragePath:(NSString*)storagePath
// ----------------------------------------------------------------------------
{
	NSArray* databaseFiles = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:storagePath error:NULL];
	for (NSString* databaseFile in databaseFiles)
	{
		if ([[databaseFile pathExtension] caseInsensitiveCompare:@"sqlitedb"] == NSOrderedSame)
			return [storagePath stringByAppendingPathComponent:databaseFile];
	}
	
	return nil;
}


// ----------------------------------------------------------------------------
- (void) createMediaFolderIfMissing:(NSString*)folderName
// ----------------------------------------------------------------------------
{
    NSString* path = [databaseStoragePath stringByAppendingPathComponent:folderName];
    
    BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:path];
    if (!exists)
        [[NSFileManager defaultManager] createDirectoryAtPath:path withIntermediateDirectories:YES attributes:nil error:NULL];
}

	
// ----------------------------------------------------------------------------
- (void) loadDatabase:(GBSourceListItem*)databaseItem
// ----------------------------------------------------------------------------
{
	[gameTableDataSource clearTable];
	[_gameDetailController hideAllFields];
	[_gameDetailController hideDatabaseDownloadView];

	_gameTableLoaded = NO;
	_databasePath = nil;
	
	databaseStoragePath = [databaseItem path];
    BOOL isDir = NO;
	BOOL databaseStorageExists = [[NSFileManager defaultManager] fileExistsAtPath:databaseStoragePath isDirectory:&isDir];
	
	if (databaseStorageExists && isDir)
    {
        [self createMediaFolderIfMissing:@"Extras"];
        [self createMediaFolderIfMissing:@"Screenshots"];
        [self createMediaFolderIfMissing:@"Games"];
		_databasePath = [GameBaseAppDelegate databasePathFromDatabaseStoragePath:databaseStoragePath];
    }
	else
	{
        NSAlert* alert = [[NSAlert alloc] init];
        [alert addButtonWithTitle:@"Locate Folder..."];
        [alert addButtonWithTitle:@"Cancel"];
        [alert setMessageText:@"This database folder cannot be found!"];
        [alert setInformativeText:@"Do you want to update the database entry by selecting a new folder?"];
        [alert setAlertStyle:NSWarningAlertStyle];
		
		NSInteger returnStatus = [alert runModal];
		
		if (returnStatus == NSAlertFirstButtonReturn)
			[self findMissingDatabasePath:databaseItem];

		_gameTableLoaded = YES;
		return;
	}
		
	BOOL databaseExists = [SQLite3Connector databaseExists:_databasePath];
	if (databaseExists)
	{
		[loadingLabel setHidden:NO];
		[_loadingProgressIndicator setHidden:NO];
		[_loadingProgressIndicator setIndeterminate:YES];
		[_loadingProgressIndicator startAnimation:self];

		//NSLog(@"Loading database from %@\n", databaseStoragePath);
		
		[NSThread detachNewThreadSelector:@selector(loadGameTableThread:) toTarget:self withObject:nil];
	}
	else 
	{
		databaseStoragePath = nil;
		_databasePath = nil;
		_gameTableLoaded = YES;
		
		[_gameDetailController showDatabaseDownloadView];
	}
}


// ----------------------------------------------------------------------------
- (void) loadCollection:(GBSourceListItem*)collectionItem
// ----------------------------------------------------------------------------
{
	if (!_gameTableLoaded)
		return;

	if (_databasePath == nil)
		return;
	
	BOOL databaseExists = [SQLite3Connector databaseExists:_databasePath];
	if (databaseExists)
	{
		GBGameCollection* collection = (GBGameCollection*) collectionItem.representedObject;
        [gameTableDataSource filterGamesFromCollection:collection keepSelection:NO];
	}
}


// ----------------------------------------------------------------------------
- (BOOL) validateMenuItem:(NSMenuItem*)item
// ----------------------------------------------------------------------------
{
	NSInteger tag = [item tag];

	if (tag == VALIDATION_TAG_PLAY_GAME)
		return _gameDetailController.currentItemGamePath != nil && gPreferences.mEmulatorToolPath != nil && _gameDetailController.hasSingleValidSelection;
	
	if (tag == VALIDATION_TAG_PLAY_MUSIC)
		return _gameDetailController.currentItemMusicPath != nil && gPreferences.mSidPlayerToolPath != nil && _gameDetailController.hasSingleValidSelection;

	if (tag == VALIDATION_TAG_REVEAL_GAME)
		return _gameDetailController.currentItemGamePath != nil;
	
	if (tag == VALIDATION_TAG_REVEAL_MUSIC)
		return _gameDetailController.currentItemMusicPath != nil;

	return YES;
}



//#define MEASURE_TIMING

// ----------------------------------------------------------------------------
- (void) loadGameTableThread:(id)object
// ----------------------------------------------------------------------------
{
	if (databaseStoragePath == nil)
		return;
	
#ifdef MEASURE_TIMING
	NSDate* start = [NSDate date];
#endif
	
	NSString* queryString = @"SELECT Games.GA_Id, Games.Name, Years.Year, Genres.Genre, PGenres.ParentGenre, Publishers.Publisher, Games.Classic, Games.Comment, Games.ScrnshotFilename, Games.SidFilename, Games.Filename \
							  FROM Games INNER JOIN Publishers ON (Publishers.PU_Id=Games.PU_Id) JOIN Years ON (Games.YE_Id=Years.YE_Id) JOIN Genres ON (Genres.GE_Id=Games.GE_Id) JOIN PGenres ON (PGenres.PG_Id=Genres.PG_Id);";

	NSMutableArray* gameTable = [_databaseConnector getData:queryString fromDatabase:_databasePath];

#ifdef MEASURE_TIMING
	NSDate* end = [NSDate date];
	NSLog(@"Querying database took %f seconds\n", [end timeIntervalSinceDate:start]);
#endif
	
	gameTableDataSource.gameTable = gameTable;
    [gameTableDataSource sortUsingSortDescriptors];
    gameTableDataSource.unfilteredGameTable = [gameTable mutableCopy];
	
	NSFileManager* fileManager = [NSFileManager defaultManager];
	NSString* gameDirectoryPath = [databaseStoragePath stringByAppendingPathComponent:@"Games"];
	NSString* screenshotDirectoryPath = [databaseStoragePath stringByAppendingPathComponent:@"Screenshots"];
	NSString* noScreenshotPath = [[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:@"noScreenshot.png"];

	BOOL gameDirectoryExists = [fileManager fileExistsAtPath:gameDirectoryPath isDirectory:NULL];
	BOOL screenshotDirectoryExists = [fileManager fileExistsAtPath:screenshotDirectoryPath isDirectory:NULL];
	
	NSMutableArray* imageBrowserData = imageBrowserDataSource.browserData;

	gameTableDataSource.availableGameCount = 0;

#ifdef MEASURE_TIMING
	start = [NSDate date];
#endif
	
	[_loadingProgressIndicator setIndeterminate:NO];
	
	NSInteger index = 0;
	NSInteger availableGameCount = 0;
    GameBaseAppDelegate* gamebaseAppDelegate = (GameBaseAppDelegate*)[NSApp delegate];
    GBSourceListDataSource* sourceListDataSource = gamebaseAppDelegate.sourceListDataSource;
    
    [sourceListDataSource clearClassicsCollection];
    
	for (NSMutableDictionary* gameRow in gameTable)
	{
		NSInteger year = [[gameRow objectForKey:@"Year"] intValue];
		NSString* yearString = (year >= 9000) ? @"19??" : [NSString stringWithFormat:@"%ld", (long)year];
		[gameRow setObject:yearString forKey:@"Year"];
		
		NSString* genreName = [gameRow objectForKey:@"Genre"];
		NSString* parentGenreName = [gameRow objectForKey:@"ParentGenre"];
		
		if ([parentGenreName length] > 0)
			genreName = [NSString stringWithFormat:@"%@ - %@", parentGenreName, genreName];
		else
			genreName = [NSString stringWithFormat:@"%@", genreName];
		
		[gameRow setObject:genreName forKey:@"Genre"];
		 
		NSString* gameFilename = [gameRow objectForKey:@"Filename"];
		if (gameDirectoryExists && [gameFilename length] > 0)
		{
			NSString* gameFilenameForwardSlashed = [gameFilename stringByReplacingOccurrencesOfString:@"\\" withString:@"/"];
			NSString* gamePath = [gameDirectoryPath stringByAppendingPathComponent:gameFilenameForwardSlashed];

			BOOL exists = [fileManager fileExistsAtPath:gamePath isDirectory:NULL];
			if (exists)
			{
				[gameRow setObject:gamePath forKey:@"FullGamePath"];
				availableGameCount++;
			}
		}

        NSNumber* classic = [gameRow objectForKey:@"Classic"];
        BOOL isClassic = [classic boolValue];

		NSString* screenshotFilename = [gameRow objectForKey:@"ScrnshotFilename"];
		if (databaseStoragePath != nil)
		{
            NSString* title = isClassic ? [NSString stringWithFormat:@"%C %@", 0x2605, [gameRow objectForKey:@"Name"]] : [gameRow objectForKey:@"Name"];
			NSString* subTitle = [NSString stringWithFormat:@"%@ %@", yearString, genreName];
			NSString* screenshotPath = noScreenshotPath;
			
			if (screenshotDirectoryExists && [screenshotFilename length] > 0)
			{
				NSString* screenshotForwardSlashed = [screenshotFilename stringByReplacingOccurrencesOfString:@"\\" withString:@"/"];
				screenshotPath = [screenshotDirectoryPath stringByAppendingPathComponent:screenshotForwardSlashed];
			}

			GBImageBrowserItem* imageBrowserItem = [[GBImageBrowserItem alloc] initWithPath:screenshotPath andTitle:title andSubTitle:subTitle];
			[imageBrowserData addObject:imageBrowserItem];
		}
        else
            NSLog(@"Missing databaseStoragePath for game: %@", gameRow);

        // Fill Classic collection
        if (isClassic)
            [sourceListDataSource addGameIdToClassicsCollection:[gameRow objectForKey:@"GA_Id"]];
        
		//[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1]];
		
		if ((index & 0x3ff) == 0)
		{
			gameTableDataSource.availableGameCount = availableGameCount;
			NSNumber* progressNumber = [NSNumber numberWithFloat:100.0f * (float) index / (float) [gameTable count]];
			[self performSelectorOnMainThread:@selector(reloadGameTable:) withObject:progressNumber waitUntilDone:NO];
		}
		index++;
	}

    imageBrowserDataSource.unfilteredBrowserData = [imageBrowserData mutableCopy];
	gameTableDataSource.availableGameCount = availableGameCount;
    
    
#ifdef MEASURE_TIMING
	end = [NSDate date];
	NSLog(@"Processing database took %f seconds\n", [end timeIntervalSinceDate:start]);
#endif
	
	[self performSelectorOnMainThread:@selector(gameTableLoadFinished:) withObject:nil waitUntilDone:NO];
}


// ----------------------------------------------------------------------------
- (void) reloadGameTable:(id)sender
// ----------------------------------------------------------------------------
{
	[[_window tableView] reloadData];
	
	[_loadingProgressIndicator setDoubleValue:[sender doubleValue]];
}


// ----------------------------------------------------------------------------
- (void) gameTableLoadFinished:(id)sender
// ----------------------------------------------------------------------------
{
	[_loadingProgressIndicator setHidden:YES];
	[loadingLabel setHidden:YES];
	
	[[_window tableView] reloadData];
	[imageBrowserDataSource reloadData];

	_gameTableLoaded = YES;

	NSInteger rowToSelect = 0;
	
	if (gPreferences.mLastViewedGameId != -1)
		rowToSelect = [gameTableDataSource findRowIndexOfGameWithGameId:gPreferences.mLastViewedGameId];
	
	NSIndexSet* indicesToSelect = [NSIndexSet indexSetWithIndex:rowToSelect];
	
	if ([[[_window tableView] selectedRowIndexes] count] == 0)
	{
		if (gPreferences.mImageGridMode)
		{
			[[imageBrowserDataSource browserView] setSelectionIndexes:indicesToSelect byExtendingSelection:NO];
			[[imageBrowserDataSource browserView] scrollIndexToVisible:rowToSelect];
		}
		else
		{
			[[_window tableView] selectRowIndexes:indicesToSelect byExtendingSelection:NO];
			[[_window tableView] scrollRowToVisible:rowToSelect];
		}
	}
}


// ----------------------------------------------------------------------------
- (void) findYoutubeLongplayForGame:(NSString*)gameName
// ----------------------------------------------------------------------------
{
	if (gameName == nil)
		return;
	
	NSString* searchString = [NSString stringWithFormat:@"C64+Longplay+-+%@", gameName];
    searchString = [searchString stringByAddingPercentEncodingWithAllowedCharacters:[NSCharacterSet URLQueryAllowedCharacterSet]];
	NSString* urlString = @"http://www.youtube.com/results?search_query=";
	urlString = [urlString stringByAppendingString:searchString];
	[[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:urlString]];
}


// ----------------------------------------------------------------------------
- (void) findMissingDatabasePath:(GBSourceListItem*)databaseItem
// ----------------------------------------------------------------------------
{
	if (databaseItem == nil)
		return;
	
	NSOpenPanel* openPanel = [NSOpenPanel openPanel];
	[openPanel setCanChooseFiles:NO];
	[openPanel setCanChooseDirectories:YES];
	
	NSString* directory = [[databaseItem path] stringByDeletingLastPathComponent];
	BOOL isFolder = NO;
	BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:directory isDirectory:&isFolder];
	if (!exists || !isFolder)
		directory = nil;
	
	[openPanel beginSheetForDirectory:directory file:nil types:nil modalForWindow:_window modalDelegate:self didEndSelector:@selector(didEndMissingDatabasePathSheet:returnCode:contextInfo:) contextInfo:(void*)databaseItem];
}


// ----------------------------------------------------------------------------
- (void)didEndMissingDatabasePathSheet:(NSOpenPanel*)openPanel returnCode:(int)returnCode contextInfo:(void*)contextInfo
// ----------------------------------------------------------------------------
{
	if (returnCode == NSModalResponseOK)
	{
        NSArray* filesToOpen = [openPanel filenames];
        NSString* path = [filesToOpen objectAtIndex:0];
		
		GBSourceListItem* databaseItem = (__bridge GBSourceListItem*) contextInfo;
		[databaseItem setPath:path];
		[databaseItem setNameFromString:[path lastPathComponent]];
		BOOL valid = [GameBaseAppDelegate containsValidDatabaseAtPath:path];
		[_sourceListDataSource updateDatabaseItem:databaseItem asExisting:YES andValid:valid];
		[_sourceListDataSource updateDatabaseListFromItems];
		
		[self loadDatabase:databaseItem];
    }	
}


#pragma mark -
#pragma mark UI Actions


// ----------------------------------------------------------------------------
- (IBAction) playGame:(id)sender;
// ----------------------------------------------------------------------------
{
	if (gPreferences.mEmulatorToolPath != nil && _gameDetailController.currentItemGamePath != nil)
	{
		CFUUIDRef uuid = CFUUIDCreate(NULL);
		NSString* directoryName = (__bridge_transfer NSString *) CFUUIDCreateString(NULL, uuid);
		CFRelease(uuid);
		
		NSString* tempPath = NSTemporaryDirectory();
		unzipDirectoryPath = [tempPath stringByAppendingPathComponent:directoryName];
		[[NSFileManager defaultManager] createDirectoryAtPath:unzipDirectoryPath withIntermediateDirectories:YES attributes:nil error:NULL];
		
		NSMutableArray* unzipArguments = [NSMutableArray arrayWithCapacity:3];
		
		[unzipArguments addObject:@"-o"];
		[unzipArguments addObject:@"-q"];
		[unzipArguments addObject:_gameDetailController.currentItemGamePath];
		
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(unzipTaskFinished:) name:NSTaskDidTerminateNotification object:nil];	
		unzipTask = [[NSTask alloc] init];
		
		[unzipTask setLaunchPath:@"/usr/bin/unzip"];
		[unzipTask setCurrentDirectoryPath:unzipDirectoryPath];
		[unzipTask setArguments:unzipArguments];
		
		[unzipTask launch];
		
		//NSLog(@"Play %@, %@\n", gameDetailController.currentItemGamePath, unzipDirectoryPath);
	}
}


// ----------------------------------------------------------------------------
- (void) unzipTaskFinished:(NSNotification*)aNotification
// ----------------------------------------------------------------------------
{
	NSTask* task = (NSTask*) [aNotification object];
	if (task != unzipTask)
		return;
	
	int result = [unzipTask terminationStatus];
	unzipTask = nil;
	
	if (result == 0)
	{
		NSString* extractedFilePath = nil;
		NSArray* extractedFiles = [[NSFileManager defaultManager] contentsOfDirectoryAtPath:unzipDirectoryPath error:NULL];
		
		NSArray* supportedExtensions = [NSArray arrayWithObjects:@"d64", @"t64", @"p00", @"prg", @"d71", @"d81", @"crt", nil];
		NSMutableArray* supportedFiles = [[NSMutableArray alloc] init];
		for (NSString* extractedFile in extractedFiles)
		{
			NSString* extension = [extractedFile pathExtension];
			
			for (NSString* supportedExtension in supportedExtensions)
			{
				if ([supportedExtension caseInsensitiveCompare:extension] == NSOrderedSame)
				{
					extractedFilePath = [unzipDirectoryPath stringByAppendingPathComponent:extractedFile];
					[supportedFiles addObject:extractedFilePath];
					break;
				}
			}
		}
		
		NSString* fileToRun = nil;
		
		if (_gameDetailController.currentItemGameToRun != nil)
			fileToRun = [unzipDirectoryPath stringByAppendingPathComponent:_gameDetailController.currentItemGameToRun];
		else
			fileToRun = [supportedFiles objectAtIndex:0];
		
		NSBundle* bundle = [NSBundle bundleWithPath:gPreferences.mEmulatorToolPath];
		gPreferences.mEmulatorType = [prefsWindowController getEmulatorTypeFromBundleIdentifier:[bundle bundleIdentifier]];
		
        BOOL isCrt = [[fileToRun pathExtension] caseInsensitiveCompare:@"crt"] == NSOrderedSame;
        
        if (([supportedFiles count] > 1 || isCrt) && (gPreferences.mEmulatorType == EMULATOR_VICE_X64 || gPreferences.mEmulatorType == EMULATOR_VICE_X64SC))
		{
			NSString* executablePath = [bundle executablePath];

            NSMutableArray* viceArguments = [NSMutableArray arrayWithCapacity:3];

            if (isCrt)
                [viceArguments addObject:@"-cartcrt"];
            else
            {
                NSString* flipListPath = [unzipDirectoryPath stringByAppendingPathComponent:@"vice.vfl"];
                NSString* flipList = @"# Vice fliplist file\n\nUNIT 8\n";
                [supportedFiles sortUsingSelector:@selector(compare:)];
                for (NSString* supportedFile in supportedFiles)
                    flipList = [flipList stringByAppendingFormat:@"%@\n", supportedFile];
                
                [flipList writeToFile:flipListPath atomically:YES encoding:NSUTF8StringEncoding error:NULL];
                [viceArguments addObject:@"-flipname"];
                [viceArguments addObject:@"vice.vfl"];
                [viceArguments addObject:@"-autostart"];
            }
			
			[viceArguments addObject:fileToRun];
			
			NSTask* viceTask = [[NSTask alloc] init];
			[viceTask setLaunchPath:executablePath];
			[viceTask setCurrentDirectoryPath:unzipDirectoryPath];
			[viceTask setArguments:viceArguments];
			[viceTask launch];
		}
		else if (gPreferences.mEmulatorType == EMULATOR_MICRO64)
        {
            NSString* executablePath = [bundle executablePath];
            
            NSMutableArray* micro64Arguments = [NSMutableArray arrayWithCapacity:3];
            [micro64Arguments addObject:[fileToRun stringWithPathRelativeTo:executablePath]];
            
            NSTask* micro64Task = [[NSTask alloc] init];
            [micro64Task setLaunchPath:executablePath];
            [micro64Task setCurrentDirectoryPath:unzipDirectoryPath];
            [micro64Task setArguments:micro64Arguments];
            [micro64Task launch];
        }
        else
		{
			[[NSWorkspace sharedWorkspace] openFile:fileToRun withApplication:gPreferences.mEmulatorToolPath];
			
			if ([supportedFiles count] > 1)
				[[NSWorkspace sharedWorkspace] openFile:unzipDirectoryPath];
		}
	}
}


// ----------------------------------------------------------------------------
- (IBAction) playMusic:(id)sender
// ----------------------------------------------------------------------------
{
	if (gPreferences.mSidPlayerToolPath != nil && _gameDetailController.currentItemMusicPath != nil)
		[[NSWorkspace sharedWorkspace] openFile:_gameDetailController.currentItemMusicPath withApplication:gPreferences.mSidPlayerToolPath andDeactivate:!gPreferences.mLaunchSidPlayerInBackground];
}


// ----------------------------------------------------------------------------
- (IBAction) revealGameInFinder:(id)sender
// ----------------------------------------------------------------------------
{
	if (_gameDetailController.currentItemGamePath == nil)
		return;
	
	NSWorkspace *workSpace = [NSWorkspace sharedWorkspace];
	[workSpace selectFile:_gameDetailController.currentItemGamePath inFileViewerRootedAtPath:@""];
}


// ----------------------------------------------------------------------------
- (IBAction) revealMusicInFinder:(id)sender
// ----------------------------------------------------------------------------
{
	if (_gameDetailController.currentItemMusicPath == nil)
		return;
	
	NSWorkspace *workSpace = [NSWorkspace sharedWorkspace];
	[workSpace selectFile:_gameDetailController.currentItemMusicPath inFileViewerRootedAtPath:@""];
}


// ----------------------------------------------------------------------------
- (IBAction) copyPathOfGameToClipboard:(id)sender
// ----------------------------------------------------------------------------
{
	if (_gameDetailController.currentItemGamePath == nil)
		return;
	
	NSPasteboard* pasteboard = [NSPasteboard generalPasteboard];
	[pasteboard declareTypes:[NSArray arrayWithObjects:NSStringPboardType, nil] owner:self];
	[pasteboard setString:_gameDetailController.currentItemGamePath forType:NSStringPboardType];
	
}


// ----------------------------------------------------------------------------
- (IBAction) copyPathOfMusicToClipboard:(id)sender
// ----------------------------------------------------------------------------
{
	if (_gameDetailController.currentItemMusicPath == nil)
		return;
	
	NSPasteboard* pasteboard = [NSPasteboard generalPasteboard];
	[pasteboard declareTypes:[NSArray arrayWithObjects:NSStringPboardType, nil] owner:self];
	[pasteboard setString:_gameDetailController.currentItemMusicPath forType:NSStringPboardType];
}


// ----------------------------------------------------------------------------
- (IBAction) copyRelativePathOfMusicToClipboard:(id)sender
// ----------------------------------------------------------------------------
{
	if (_gameDetailController.currentItemMusicPath == nil)
		return;
	
	if (databaseStoragePath != nil)
	{
		NSString* musicDirectoryName = @"C64Music";
		NSString* musicPath = [databaseStoragePath stringByAppendingPathComponent:musicDirectoryName];
		NSString* commonPrefix = [NSString commonRootPathOfFilename:musicPath andFilename:_gameDetailController.currentItemMusicPath];
		NSString* relativePath = [_gameDetailController.currentItemMusicPath stringByRemovingPrefix:commonPrefix];
		if ([relativePath characterAtIndex:0] == '/')
			relativePath = [relativePath stringByRemovingPrefix:@"/"];
		
		NSPasteboard* pasteboard = [NSPasteboard generalPasteboard];
		[pasteboard declareTypes:[NSArray arrayWithObjects:NSStringPboardType, nil] owner:self];
		[pasteboard setString:relativePath forType:NSStringPboardType];
	}
}


// ----------------------------------------------------------------------------
- (IBAction) moveFocusToSearchField:(id)sender
// ----------------------------------------------------------------------------
{
	[_window makeFirstResponder:[gameTableDataSource toolbarSearchField]];
}


// ----------------------------------------------------------------------------
- (void) viewModeChanged:(id)sender
// ----------------------------------------------------------------------------
{
	if ([sender isSelectedForSegment:0])
		[self switchToListMode:self];
	else
		[self switchToImageGridMode:self];
}


// ----------------------------------------------------------------------------
- (void) switchToListMode:(id)sender
// ----------------------------------------------------------------------------
{
	if ([browserView superview] != nil)
		return;
	
	[viewModeSwitch setSelectedSegment:0];

	NSRect frame = [imageBrowserView frame];
	[imageBrowserView removeFromSuperview];
	[browserView setFrame:frame];
	[dataContentView addSubview:browserView];
	[zoomSliderView setHidden:YES];
	[_window makeFirstResponder:browserView];

	gPreferences.mImageGridMode = NO;
}


// ----------------------------------------------------------------------------
- (void) switchToImageGridMode:(id)sender
// ----------------------------------------------------------------------------
{
	if ([imageBrowserView superview] != nil)
		return;
	
	[viewModeSwitch setSelectedSegment:1];
	
	NSRect frame = [browserView frame];
	[browserView removeFromSuperview];
	[imageBrowserView setFrame:frame];
	[dataContentView addSubview:imageBrowserView];
	[zoomSliderView setHidden:NO];
	
	[imageBrowserDataSource reloadData];
	NSIndexSet* selectedIndices = [gameTableDataSource.gameTableView selectedRowIndexes];
	[[imageBrowserDataSource browserView] setSelectionIndexes:selectedIndices byExtendingSelection:NO];
	[[imageBrowserDataSource browserView] scrollIndexToVisible:[selectedIndices firstIndex]];
	
	[_window makeFirstResponder:imageBrowserView];
	
	gPreferences.mImageGridMode = YES;
}


// ----------------------------------------------------------------------------
- (IBAction) showPreferencesWindow:(id)sender
// ----------------------------------------------------------------------------
{
	if (prefsWindowController == nil)
		prefsWindowController = [[GBPreferencesWindowController alloc] init];
	
	[prefsWindowController showWindow:sender];
}



#pragma mark -
#pragma mark split view delegate methods


// ----------------------------------------------------------------------------
- (BOOL) splitView:(NSSplitView*)splitView canCollapseSubview:(NSView*)subview
// ----------------------------------------------------------------------------
{
    return NO;
}



// ----------------------------------------------------------------------------
- (CGFloat) splitView:(NSSplitView*)splitView constrainSplitPosition:(CGFloat)proposedPosition ofSubviewAt:(NSInteger)offset
// ----------------------------------------------------------------------------
{
	if (splitView == sourceListSplitView && offset == 0)
	{
		float position = fminf(proposedPosition, 220.0f);
		position = fmaxf(100.0f, position);
		return position;
	}

	return proposedPosition;
}


// ----------------------------------------------------------------------------
- (NSRect) splitView:(NSSplitView *)splitView effectiveRect:(NSRect)proposedEffectiveRect forDrawnRect:(NSRect)drawnRect ofDividerAtIndex:(NSInteger)dividerIndex
// ----------------------------------------------------------------------------
{
    return NSInsetRect(proposedEffectiveRect, -2.0f, 0.0f);

//	if (splitView == sourceListSplitView)
//		return NSInsetRect(proposedEffectiveRect, -2.0f, 0.0f);
//	else
//		return proposedEffectiveRect;
}


// ----------------------------------------------------------------------------
- (BOOL) splitView:(NSSplitView *)splitView shouldAdjustSizeOfSubview:(NSView *)subview
// ----------------------------------------------------------------------------
{
    return NO;
}


// ----------------------------------------------------------------------------
- (id)validRequestorForSendType:(NSString *)sendType returnType:(NSString *)returnType
// ----------------------------------------------------------------------------
{
	return nil;
}


@end


@implementation GBButtonToolbarItem

// ----------------------------------------------------------------------------
- (void) validate
// ----------------------------------------------------------------------------
{
    NSInteger tag = [self tag];
    
    GameBaseAppDelegate* gamebaseAppDelegate = (GameBaseAppDelegate*)[NSApp delegate];
    BOOL valid = NO;
    
    if (tag == VALIDATION_TAG_PLAY_GAME)
        valid = gamebaseAppDelegate.gameDetailController.currentItemGamePath != nil && gPreferences.mEmulatorToolPath != nil && gamebaseAppDelegate.gameDetailController.hasSingleValidSelection;
    else if (tag == VALIDATION_TAG_PLAY_MUSIC)
        valid = gamebaseAppDelegate.gameDetailController.currentItemMusicPath != nil && gPreferences.mSidPlayerToolPath != nil && gamebaseAppDelegate.gameDetailController.hasSingleValidSelection;
    else if (tag == VALIDATION_TAG_FIND_VIDEO)
        valid = gamebaseAppDelegate.gameDetailController.hasSingleValidSelection;

    [(NSButton*)self.view setEnabled:valid];
}

@end